"""
background_systems.py contains BackgroundSystem, which contains different arrays relating to
the evolution of fuel blends, sulfur content in fuels, and electricity mixes in different
countries over time.
"""


from typing import Dict

import numpy as np
import pandas as pd
import xarray as xr

from . import DATA_DIR


def data_to_dict(csv_list: list) -> dict:
    """
    Returns a dictionary from a sequence of items.
    :param csv_list: list
    :return: dict
    """

    (_, *header), *data = csv_list
    csv_dict = {}
    for row in data:
        key, *values = row
        csv_dict[key] = dict(zip(header, values))

    return csv_dict


def get_electricity_losses() -> Dict[str, float]:
    """
    Retrieve cumulative electricity losses from high to medium and low voltage.
    Source: `ecoinvent v.3.6 <https://www.ecoinvent.org/>`_.

    :returns: dictionary
    :rtype: dict

    """
    filename = "cumulative_electricity_losses.csv"
    filepath = DATA_DIR / filename
    if not filepath.is_file():
        raise FileNotFoundError(
            "The CSV file that contains electricity mixes could not be found."
        )
    with open(filepath, encoding="utf-8") as file:
        csv_list = [[val.strip() for val in r.split(";")] for r in file.readlines()]

    return data_to_dict(csv_list)


def get_region_mapping() -> Dict[str, str]:
    """
    Retrieve mapping between ISO country codes and REMIND regions.

    :returns: dictionary
    :rtype: dict

    """
    filename = "region_mapping.csv"
    filepath = DATA_DIR / filename
    if not filepath.is_file():
        raise FileNotFoundError(
            "The CSV file that contains correspondences between REMIND region "
            "names and ISO country codes could not be found."
        )
    with open(filepath, encoding="utf-8") as file:
        csv_list = [[val.strip() for val in r.split(";")] for r in file.readlines()]

    return data_to_dict(csv_list)


def get_electricity_mix() -> xr.DataArray:
    """
    Retrieve electricity mixes and shape them into a xarray.
    Source:
        * for European countries (`ENTSOE TYNDP 2020 scenarios <https://2020.entsos-tyndp-scenarios.eu/>`_),
        * for African countries (`TEMBA <http://www.osemosys.org/temba.html>`_ model)
        * and for other countries (`IEA World Energy outlook 2017 <https://www.iea.org/reports/world-energy-outlook-2017>`_)

    :returns: An axarray with 'country' and 'year' as dimensions
    :rtype: xarray.core.dataarray.DataArray

    """
    filename = "electricity_mixes.csv"
    filepath = DATA_DIR / filename
    if not filepath.is_file():
        raise FileNotFoundError(
            "The CSV file that contains electricity mixes could not be found."
        )

    dataframe = pd.read_csv(filepath, sep=";", index_col=["country", "year"])
    dataframe = dataframe.reset_index()

    array = (
        dataframe.melt(id_vars=["country", "year"], value_name="value")
        .groupby(["country", "year", "variable"])["value"]
        .mean()
        .to_xarray()
    )
    array = array.interpolate_na(
        dim="year", method="linear", fill_value="extrapolate"
    ).clip(0, 1)
    array /= array.sum(axis=2)

    return array


def get_biofuel_share() -> xr.DataArray:
    """
    Retrieve shares of biofuel consumption from REMIND and shape them into an xarray.

    :return: An axarray with 'country' and 'year' as dimensions
    :rtype: xarray.core.dataarray.DataArray
    """
    filename = "biofuel_share.csv"
    filepath = DATA_DIR / filename

    if not filepath.is_file():
        raise FileNotFoundError(
            "The CSV file that contains biofuel shares could not be found."
        )
    dataframe = pd.read_csv(filepath, sep=";")

    country_codes = dataframe["Region"].unique()
    scenarios = dataframe["Scenario"].unique()
    year = dataframe["Year"].unique()
    tech = dataframe.columns[3:]
    array = xr.DataArray(
        np.zeros((len(country_codes), len(year), len(scenarios), len(tech), 1)),
        coords=[country_codes, year, scenarios, tech, np.arange(1)],
        dims=["region", "year", "scenario", "fuel_type", "value"],
    )
    for country_code in country_codes:
        for scenario in scenarios:
            val = dataframe.loc[
                (dataframe["Region"] == country_code)
                & (dataframe["Scenario"] == scenario),
                "Biomass fuel":,
            ]
            array.loc[dict(region=country_code, scenario=scenario, value=0)] = val
    return array


def get_biomethane_share() -> xr.DataArray:
    """

    :return: Returns an xarray with share of biomethane in the fuel blend, per country, per year.
    """
    filename = "share_bio_cng.csv"
    filepath = DATA_DIR / filename

    if not filepath.is_file():
        raise FileNotFoundError(
            "The CSV file that contains biomethane shares could not be found."
        )
    dataframe = pd.read_csv(filepath, sep=";")

    return dataframe.groupby(["country", "year"]).sum().to_xarray().to_array()


def get_biogasoline_share() -> xr.DataArray:
    """

    :return: Returns an xarray with share of bioethanol in the fuel blend, per country, per year.
    """
    filename = "share_bio_gasoline.csv"
    filepath = DATA_DIR / filename

    if not filepath.is_file():
        raise FileNotFoundError(
            "The CSV file that contains biogasoline shares could not be found."
        )
    dataframe = pd.read_csv(filepath, sep=";")

    return dataframe.groupby(["country", "year"]).sum().to_xarray().to_array()


def get_biodiesel_share() -> xr.DataArray:
    """

    :return: Returns an xarray with share of biodiesel in the fuel blend, per country, per year.
    """
    filename = "share_bio_diesel.csv"
    filepath = DATA_DIR / filename

    if not filepath.is_file():
        raise FileNotFoundError(
            "The CSV file that contains biodiesel shares could not be found."
        )
    dataframe = pd.read_csv(filepath, sep=";")

    return dataframe.groupby(["country", "year"]).sum().to_xarray().to_array()


def get_sulfur_content_in_fuel() -> xr.DataArray:
    """
    Retrieve sulfur content per kg of petrol and diesel.
    For CH, DE, FR, AU and SE, the concentration values come from HBEFA 4.1, from 1909 to 2020 (extrapolated to 2050).

    For the other countries, values come from
    Miller, J. D., & Jin, L. (2019). Global progress toward soot-free diesel vehicles in 2019.
    International Council on Clean Transportation.
    https://www.theicct.org/publications/global-progress-toward-soot-free-diesel-vehicles-2019

    There is an assumption made: countries that have high-sulfur content fuels (above 50 ppm in 2019) are assumed to
    improve over time to reach 50 ppm by 2050.

    :return: An axarray with 'country' and 'year' as dimensions
    :rtype: xarray.core.dataarray.DataArray
    """
    filename = "S_concentration_fuel.csv"
    filepath = DATA_DIR / filename

    if not filepath.is_file():
        raise FileNotFoundError(
            "The CSV file that contains sulfur concentration values could not be found."
        )
    dataframe = pd.read_csv(filepath, sep=";")
    dataframe = dataframe.groupby(["country", "year"]).sum().unstack()
    dataframe.loc[:, ("diesel", 1990)] = dataframe["diesel"].max(1)
    dataframe.loc[:, ("petrol", 1990)] = dataframe["petrol"].max(1)

    dataframe.loc[dataframe[("diesel", 2019)] > 50 / 1e6, ("diesel", 2050)] = 50 / 1e6
    dataframe.loc[dataframe[("petrol", 2019)] > 50 / 1e6, ("petrol", 2050)] = 50 / 1e6
    dataframe.loc[:, ("diesel", 2050)] = dataframe["diesel"].min(1)
    dataframe.loc[:, ("petrol", 2050)] = dataframe["petrol"].min(1)

    dataframe = dataframe.interpolate(axis=1)
    dataframe = dataframe.unstack().reset_index()
    dataframe = dataframe.rename(columns={"level_0": "fuel"})
    arr = dataframe.groupby(["country", "year", "fuel"]).sum()[0].to_xarray()
    return arr


class BackgroundSystemModel:
    """
    Retrieve and build dictionaries that contain important information to model in the background system:

        * gross electricity production mixes from nearly all countries in the world, from 2015 to 2050.
        * cumulative electricity transformation/transmission/distribution losses from high voltage to medium and low voltage.
        * share of biomass-derived fuel in the total consumption of liquid fuel in the transport sector. Source: REMIND.
        * share of bioethanol, biodiesel and biomethane, for each country, for different years.
        * share of sulfur in gasoline and diesel, for different countries and years.
    """

    def __init__(self) -> None:
        self.electricity_mix = get_electricity_mix()
        self.losses = get_electricity_losses()
        self.region_map = get_region_mapping()
        self.biofuel = get_biofuel_share()
        self.sulfur = get_sulfur_content_in_fuel()
        self.biomethane = get_biomethane_share()
        self.biogasoline = get_biogasoline_share()
        self.biodiesel = get_biodiesel_share()

    def __str__(self):
        return self.__class__.__name__



"""
car_input_parameters.py contains the CarInputParameters class, which, after instantiated, contains
all definitions (metadata) of input and calculated parameters, along with their values.
Also, it inherits methods from `klausen`, which exposes the methods .static() and .stochastic(),
which generate single or random values for input parameters.
"""

import json
from pathlib import Path
from typing import Union

from klausen import NamedParameters

DEFAULT = Path(__file__, "..").resolve() / "data" / "default_parameters.json"
EXTRA = Path(__file__, "..").resolve() / "data" / "extra_parameters.json"


def load_parameters(obj: Union[str, Path, list]) -> list:
    """
    Returns a json object containing parameters' definitions
    :param obj: A filepath to a json file, or a json object
    :return: Returns a json object containing parameters' definitions
    """
    if isinstance(obj, (str, Path)):
        assert Path(obj).exists(), "Can't find this filepath"
        return json.load(open(obj, encoding="utf-8"))

    # Already in correct form, just return
    return obj


class CarInputParameters(NamedParameters):
    """
    A class used to represent vehicles with associated type, size, technology, year and parameters.

    This class inherits from NamedParameters, located in the *klausen* package.
    It sources default parameters for all vehicle types from a dictionary in
    default_parameters and format them into an array following the structured described
    in the *klausen* package.

    :ivar sizes: List of string items e.g., ['Large', 'Lower medium', 'Medium', 'Mini', 'Medium SUV', 'Small', 'Van']
    :ivar powertrains: List of string items e.g., ['BEV', 'FCEV', 'HEV-p', 'ICEV-d', 'ICEV-g', 'ICEV-p', 'PHEV-c', 'PHEV-e']
    :ivar parameters: json or filepath to json containing input parameter's definitions
    :ivar extra: json or filepath to json containing calculated parameter's definitions
    :ivar years: List of integers e.g., [2000, 2010, 2020, 2040]
    :ivar metadata: Dictionary for metadata.
    :ivar values: Dictionary for storing values, of format {'param':[value]}.
    :ivar iterations: Number of iterations executed by the method :func:`~car_input_parameters.CarInputParameters.stochastic`.
        None if :func:`~car_input_parameters.CarInputParameters.static` used instead.
    :vartype iterations: int


    """

    def __init__(
        self,
        parameters: Union[str, Path, list] = None,
        extra: Union[str, Path, list] = None,
    ) -> None:
        """Create a `klausen <https://github.com/cmutel/klausen>`__ model with the car input parameters."""
        super().__init__(None)

        parameters = load_parameters(DEFAULT if parameters is None else parameters)
        extra = set(load_parameters(EXTRA if extra is None else extra))

        if not isinstance(parameters, dict):
            raise ValueError(
                f"Parameters are not correct type (expected `dict`, got `{type(parameters)}`)"
            )
        if not isinstance(extra, set):
            raise ValueError(
                f"Extra parameters are not correct type (expected `set`, got `{type(extra)}`)"
            )
        self.sizes = sorted(
            {size for o in parameters.values() for size in o.get("sizes", [])}
        )
        self.powertrains = sorted(
            {pt for o in parameters.values() for pt in o.get("powertrain", [])}
        )
        self.parameters = sorted(
            {o["name"] for o in parameters.values()}.union(set(extra))
        )

        # keep a list of input parameters, for sensitivity purpose
        self.input_parameters = sorted({o["name"] for o in parameters.values()})

        self.years = sorted({o["year"] for o in parameters.values()})
        self.add_car_parameters(parameters)

    def add_car_parameters(self, parameters: dict) -> None:
        """
        Split data and metadata according to ``klausen`` convention.

        The parameters are split into the *metadata* and *values* attributes
        of the CarInputParameters class by the add_parameters() method of the parent class.

        :param parameters: A dictionary that contains parameters.
        :type parameters: dict


        """
        keys = {"kind", "uncertainty_type", "amount", "loc", "minimum", "maximum"}

        reformatted = {}
        for key, dct in parameters.items():
            reformatted[key] = {k: v for k, v in dct.items() if k in keys}
            reformatted[key]["metadata"] = {
                k: v for k, v in dct.items() if k not in keys
            }

        self.add_parameters(reformatted)



"""
Module array.py exposes fill_xarray_from_input_parameters() which formats
parameters contained in the CarInputParameters object into a multi-dimensional array.
This array host input as well as not yet calculated parameters.
Also, modify_xarray_from_custom_parameters() offers a way to modify some of the default values
used for input parameters.
"""

from typing import Tuple, Union

import numpy as np
import pandas as pd
import stats_arrays as sa
import xarray as xr

from .car_input_parameters import CarInputParameters as c_i_p


def fill_xarray_from_input_parameters(
    cip: c_i_p, sensitivity: bool = False, scope: dict = None
) -> Tuple[Tuple[dict, dict, dict, dict], xr.DataArray]:

    """Create an `xarray` labeled array from the sampled input parameters.

    This function extracts the parameters' names and values contained in the
    `parameters` attribute of the :class:`CarInputParameters` class in
    :mod:`car_input_parameters` and insert them into a
    multi-dimensional numpy-like array from the *xarray* package
    (http://xarray.pydata.org/en/stable/).


    :param cip: Instance of the :class:`CarInputParameters` class in :mod:`car_input_parameters`.
    :param sensitivity: boolean. Whether a sensitivity test is carried out.
    :param scope: a dictionary to narrow down the scope of vehicles to consider
    :returns: `tuple`, `xarray.DataArray`
    - tuple (`size_dict`, `powertrain_dict`, `parameter_dict`, `year_dict`)
    - array

    Dimensions of `array`:

        0. Vehicle size, e.g. "small", "medium". str.
        1. Powertrain, e.g. "ICE-p", "BEV". str.
        2. Year. int.
        3. Samples.

    """

    # Check whether the argument passed is a cip object
    if not isinstance(cip, c_i_p):
        raise TypeError(
            "The argument passed is not an object of the CarInputParameter class"
        )

    if scope is None:
        scope = {"size": cip.sizes, "powertrain": cip.powertrains, "year": cip.years}
    else:
        if "size" not in scope:
            scope["size"] = cip.sizes
        if "powertrain" not in scope:
            scope["powertrain"] = cip.powertrains
        if "year" not in scope:
            scope["year"] = cip.years

    # Make sure to include PHEV-e and PHEV-c-p/d if
    # PHEV-d or PHEV-p are listed

    if "PHEV-p" in scope["powertrain"]:
        for pwt in ["PHEV-e", "PHEV-c-p"]:
            if pwt not in scope["powertrain"]:
                scope["powertrain"].append(pwt)

    if "PHEV-d" in scope["powertrain"]:
        for pwt in ["PHEV-e", "PHEV-c-d"]:
            if pwt not in scope["powertrain"]:
                scope["powertrain"].append(pwt)

    if any(size for size in scope["size"] if size not in cip.sizes):
        raise ValueError("One of the size types is not valid.")

    if any(y for y in scope["year"] if y not in cip.years):
        raise ValueError("One of the years defined is not valid.")

    if any(pwt for pwt in scope["powertrain"] if pwt not in cip.powertrains):
        raise ValueError("One of the powertrain types is not valid.")

    if not sensitivity:
        array = xr.DataArray(
            np.zeros(
                (
                    len(scope["size"]),
                    len(scope["powertrain"]),
                    len(cip.parameters),
                    len(scope["year"]),
                    cip.iterations or 1,
                )
            ).astype("float32"),
            coords=[
                scope["size"],
                scope["powertrain"],
                cip.parameters,
                scope["year"],
                np.arange(cip.iterations or 1),
            ],
            dims=["size", "powertrain", "parameter", "year", "value"],
        )
    else:
        params = ["reference"]
        params.extend(cip.input_parameters)
        array = xr.DataArray(
            np.zeros(
                (
                    len(scope["size"]),
                    len(scope["powertrain"]),
                    len(cip.parameters),
                    len(scope["year"]),
                    len(params),
                )
            ).astype("float32"),
            coords=[
                scope["size"],
                scope["powertrain"],
                cip.parameters,
                scope["year"],
                params,
            ],
            dims=["size", "powertrain", "parameter", "year", "value"],
        )

    size_dict = {k: i for i, k in enumerate(scope["size"])}
    powertrain_dict = {k: i for i, k in enumerate(scope["powertrain"])}
    year_dict = {k: i for i, k in enumerate(scope["year"])}
    parameter_dict = {k: i for i, k in enumerate(cip.parameters)}

    if not sensitivity:

        for param in cip:

            pwt = (
                set(cip.metadata[param]["powertrain"])
                if isinstance(cip.metadata[param]["powertrain"], list)
                else {cip.metadata[param]["powertrain"]}
            )

            size = (
                set(cip.metadata[param]["sizes"])
                if isinstance(cip.metadata[param]["sizes"], list)
                else {cip.metadata[param]["sizes"]}
            )

            year = (
                set(cip.metadata[param]["year"])
                if isinstance(cip.metadata[param]["year"], list)
                else {cip.metadata[param]["year"]}
            )

            if (
                pwt.intersection(scope["powertrain"])
                and size.intersection(scope["size"])
                and year.intersection(scope["year"])
            ):

                array.loc[
                    dict(
                        powertrain=[p for p in pwt if p in scope["powertrain"]],
                        size=[s for s in size if s in scope["size"]],
                        year=[y for y in year if y in scope["year"]],
                        parameter=cip.metadata[param]["name"],
                    )
                ] = cip.values[param]
    else:

        for param in cip.input_parameters:

            names = [n for n in cip.metadata if cip.metadata[n]["name"] == param]

            for name in names:

                pwt = (
                    set(cip.metadata[name]["powertrain"])
                    if isinstance(cip.metadata[name]["powertrain"], list)
                    else {cip.metadata[name]["powertrain"]}
                )

                size = (
                    set(cip.metadata[name]["sizes"])
                    if isinstance(cip.metadata[name]["sizes"], list)
                    else {cip.metadata[name]["sizes"]}
                )

                year = (
                    set(cip.metadata[name]["year"])
                    if isinstance(cip.metadata[name]["year"], list)
                    else {cip.metadata[name]["year"]}
                )

                vals = [
                    cip.values[name] for _ in range(0, len(cip.input_parameters) + 1)
                ]
                vals[cip.input_parameters.index(param) + 1] *= 1.1

                array.loc[
                    dict(
                        powertrain=[p for p in pwt if p in scope["powertrain"]],
                        size=[s for s in size if s in scope["size"]],
                        year=[y for y in year if y in scope["year"]],
                        parameter=cip.metadata[name]["name"],
                    )
                ] = vals

    return (size_dict, powertrain_dict, parameter_dict, year_dict), array


def modify_xarray_from_custom_parameters(
    filepath: Union[dict, str], array: xr.DataArray
) -> None:
    """
    Override default parameters values in `xarray` based on values provided by the user.

    This function allows to override one or several default parameter values by providing either:

        * a file path to an Excel workbook that contains the new values
        * or a dictionary

    The dictionary must be of the following format:

    .. code-block:: python

            {
                (parameter category,
                    powertrain,
                    size,
                    parameter name,
                    uncertainty type): {
                                        (year, 'loc'): value,
                                        (year, 'scale'): value,
                                        (year, 'shape'): value,
                                        (year, 'minimum'): value,
                                        (year, 'maximum'): value
                }

            }

    For example:

    .. code-block:: python

            {
                ('Driving',
                'all',
                'all',
                'lifetime kilometers',
                'none'): {
                    (2018, 'loc'): 150000, (2040, 'loc'): 150000
                    }

            }

    :param filepath: File path of workbook with new values or dictionary.
    :type filepath: str or dict
    :param array: array with vehicle parameters

    """

    if isinstance(filepath, str):
        try:
            param_dictionary = pd.read_excel(
                filepath,
                header=[0, 1],
                index_col=[0, 1, 2, 3, 4],
                sheet_name="Custom_parameters",
            ).to_dict(orient="index")
        except FileNotFoundError as err:
            raise FileNotFoundError("Custom parameters file not found.") from err
    elif isinstance(filepath, dict):
        param_dictionary = filepath
    else:
        raise TypeError("The format passed as parameter is not valid.")

    forbidden_keys = ["Driving cycle", "Background", "Functional unit"]

    for key in param_dictionary:
        if key[0] not in forbidden_keys:
            if not isinstance(key[1], str):
                powertrains = [p.strip() for p in key[1] if p]
                powertrains = [p for p in powertrains if p]
                powertrains = list(powertrains)
            elif key[1] == "all":
                powertrains = array.coords["powertrain"].values
            else:
                if key[1] in array.coords["powertrain"].values:
                    powertrains = [key[1]]
                elif all(
                    p
                    for p in key[1].split(", ")
                    if p in array.coords["powertrain"].values
                ):
                    powertrains = key[1].split(", ")
                else:
                    print(
                        f"{key[1]} is not a recognized powertrain. It will be skipped."
                    )
                    continue

            if not isinstance(key[2], str):
                sizes = [s.strip() for s in key[2] if s]
                sizes = [s for s in sizes if s]
                sizes = list(sizes)
            elif key[2] == "all":
                sizes = array.coords["size"].values
            else:
                if key[2] in array.coords["size"].values:
                    sizes = [key[2]]
                elif all(
                    s for s in key[2].split(", ") if s in array.coords["size"].values
                ):
                    sizes = key[2].split(", ")
                else:
                    print(
                        f"{key[2]} is not a recognized size category. It will be skipped."
                    )
                    continue

            param = key[3]

            if not param in array.coords["parameter"].values:
                print(f"{param} is not a recognized parameter. It will be skipped.")
                continue

            val = param_dictionary[key]

            distr_dic = {
                "triangular": 5,
                "lognormal": 2,
                "normal": 3,
                "uniform": 4,
                "none": 1,
            }
            distr = distr_dic[key[4]]

            years = {v[0] for v in val}

            for year in years:
                # No uncertainty parameters given
                if distr == 1:
                    # There should be at least a `loc`
                    if ~np.isnan(val[(year, "loc")]):
                        for size in sizes:
                            for pwt in powertrains:
                                array.loc[
                                    dict(
                                        powertrain=pwt,
                                        size=size,
                                        year=year,
                                        parameter=param,
                                    )
                                ] = val[(year, "loc")]
                    # Otherwise warn
                    else:
                        print(f"`loc`parameter missing for {param} in {year}.")
                        continue

                elif distr in [2, 3, 4, 5]:

                    # Check if the correct parameters are present
                    # Triangular

                    if distr == 5:
                        if (
                            np.isnan(val[(year, "loc")])
                            or np.isnan(val[(year, "minimum")])
                            or np.isnan(val[(year, "maximum")])
                        ):
                            print(
                                "One or more parameters for the triangular distribution "
                                f"is/are missing for {param} in {year}.\n "
                                "The parameter is skipped and default value applies."
                            )
                            continue

                    # Lognormal
                    if distr == 2:
                        if np.isnan(val[(year, "loc")]) or np.isnan(
                            val[(year, "scale")]
                        ):
                            print(
                                "One or more parameters for the lognormal distribution "
                                f"is/are missing for {param} in {year}.\n "
                                "The parameter is skipped and default value applies"
                            )
                            continue

                    # Normal
                    if distr == 3:
                        if np.isnan(val[(year, "loc")]) or np.isnan(
                            val[(year, "scale")]
                        ):
                            print(
                                "One or more parameters for the normal distribution is/are missing"
                                f" for {param} in {year}.\n"
                                f"The parameter is skipped and default value applies"
                            )
                            continue

                    # Uniform
                    if distr == 4:
                        if np.isnan(val[(year, "minimum")]) or np.isnan(
                            val[(year, "maximum")]
                        ):
                            print(
                                "One or more parameters for the uniform distribution "
                                f"is/are missing for {param} in {year}.\n "
                                "The parameter is skipped and default value applies"
                            )
                            continue

                    uncertainty_params = sa.UncertaintyBase.from_dicts(
                        {
                            "loc": val[year, "loc"],
                            "scale": val[year, "scale"],
                            "shape": val[year, "shape"],
                            "minimum": val[year, "minimum"],
                            "maximum": val[year, "maximum"],
                            "uncertainty_type": distr,
                        }
                    )

                    # Stochastic mode
                    if array.sizes["value"] > 1:

                        rng = sa.MCRandomNumberGenerator(uncertainty_params)

                        for size in sizes:
                            for pwt in powertrains:
                                array.loc[
                                    dict(
                                        powertrain=pwt,
                                        size=size,
                                        year=year,
                                        parameter=param,
                                    )
                                ] = rng.generate(array.sizes["value"]).reshape((-1,))
                    else:

                        dist = sa.uncertainty_choices[distr]
                        median = float(dist.ppf(uncertainty_params, np.array((0.5,))))

                        for size in sizes:
                            for pwt in powertrains:
                                array.loc[
                                    dict(
                                        powertrain=pwt,
                                        size=size,
                                        year=year,
                                        parameter=param,
                                    )
                                ] = median

                else:
                    print(
                        f"The uncertainty type is not recognized for {param} in {year}.\n"
                        f"The parameter is skipped and default value applies"
                    )
                    continue



from typing import Dict, List, Tuple, Union

import numexpr as ne
import numpy as np
import xarray as xr
import yaml

from . import DATA_DIR
from .energy_consumption import EnergyConsumptionModel
from .hot_emissions import HotEmissionsModel
from .noise_emissions import NoiseEmissionsModel
from .particulates_emissions import ParticulatesEmissionsModel


def finite(array, mask_value=0):
    return np.where(np.isfinite(array), array, mask_value)


class CarModel:

    """
    This class represents the entirety of the vehicles considered, with useful attributes, such as an array that stores
    all the vehicles parameters.

    :ivar array: multi-dimensional numpy-like array that contains parameters' value(s)
    :ivar cycle: name of a driving cycle, or custom driving cycle
    :ivar gradient: series of gradients, for each second of the driving cycle
    :ivar energy_storage: dictionary with selection of battery chemistry for each powertrain

    """

    def __init__(
        self,
        array: xr.DataArray,
        cycle: Union[None, str, np.ndarray] = None,
        gradient: Union[None, np.ndarray] = None,
        energy_storage: Union[None, Dict] = None,
    ) -> None:

        self.array = array

        if cycle is None:
            self.ecm = EnergyConsumptionModel("WLTC")
        else:
            self.ecm = EnergyConsumptionModel(cycle=cycle, gradient=gradient)

        self.energy_storage = None
        self.set_battery_preferences(energy_storage)
        self.energy = None

    def __call__(self, key: Union[str, List]):

        """
        This method fixes a dimension of the `array` attribute given a powertrain technology selected.
        Set up this class as a context manager, so we can have some nice syntax

        .. code-block:: python

            with class('some powertrain') as cpm:
                cpm['something']. # Will be filtered for the correct powertrain

        On with block exit, this filter is cleared
        https://stackoverflow.com/a/10252925/164864

        :param key: A powertrain type, e.g., "FCEV"
        :return: An instance of `array` filtered after the powertrain selected.

        """
        if isinstance(key, str):
            key = [key]

        self.__cache = self.array
        self.array = self.array.loc[
            dict(powertrain=[k for k in key if k in self.array.powertrain])
        ]
        return self

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.array = self.__cache
        del self.__cache

    def __getitem__(self, key: Union[str, List]) -> xr.DataArray:
        """
        Make class['foo'] automatically filter for the parameter 'foo'
        Makes the model code much cleaner

        :param key: Parameter name
        :return: `array` filtered after the parameter selected
        """

        return self.array.loc[dict(parameter=key)]

    def __setitem__(self, key, value):
        self.array.loc[{"parameter": key}] = value

    def set_all(self, drop_hybrids: bool = True, electric_utility_factor: float = None):
        """
        This method runs a series of other methods to obtain the tank-to-wheel energy requirement, efficiency
        of the car, costs, etc.

        :meth:`set_component_masses()`, :meth:`set_car_masses()` and :meth:`set_power_parameters()` are interdependent.
        `powertrain_mass` depends on `power`, `curb_mass` is affected by changes in `powertrain_mass`,
        `combustion engine mass` and `electric engine mass`, and `power` is a function of `curb_mass`.
        The current solution is to loop through the methods until the increment in driving mass is
        inferior to 0.1%.

        :param drop_hybrids: boolean. True by default. If False, the underlying vehicles used to build plugin-hybrid
                vehicles remain present in the array.
        :param electric_utility_factor: array. If an array is passed, its values are used to override the
                electric utility factor for plugin hybrid vehicles. If not, this factor is calculated using a relation
                described in `set_electric_utility_factor()`

        :returns: Does not return anything. Modifies ``self.array`` in place.

        """

        diff = 1.0

        while diff > 0.001:
            old_driving_mass = self["driving mass"].sum().values

            self.set_car_masses()

            self.set_power_parameters()
            self.set_component_masses()
            self.set_battery_properties()
            self.set_battery_fuel_cell_replacements()
            self.set_recuperation()
            self.set_fuel_cell_parameters()
            self.set_energy_stored_properties()

            diff = (self["driving mass"].sum().values - old_driving_mass) / self[
                "driving mass"
            ].sum()

        self.set_auxiliaries()
        self.set_ttw_efficiency()
        self.calculate_ttw_energy()
        self.set_share_recuperated_energy()
        self.adjust_cost()
        self.set_range()
        self.set_electric_utility_factor(electric_utility_factor)
        self.set_electricity_consumption()
        self.set_costs()
        self.set_hot_emissions()
        self.set_particulates_emission()
        self.set_noise_emissions()
        self.create_PHEV()
        if drop_hybrids:
            self.drop_hybrid()

        # we flag cars that have a range inferior to 100 km
        # and also BEVs, PHEVs and FCEVs from before 2013
        self.array.loc[dict(parameter="has_low_range")] = (
            self.array.loc[dict(parameter="range")] < 100
        )

        self.array.loc[
            dict(
                parameter="has_low_range",
                powertrain=[
                    pt
                    for pt in [
                        "BEV",
                        "PHEV-e",
                        "PHEV-c-p",
                        "PHEV-c-d",
                        "FCEV",
                        "PHEV-p",
                        "PHEV-d",
                        "HEV-d",
                        "HEV-p",
                    ]
                    if pt in self.array.coords["powertrain"].values
                ],
                year=[y for y in self.array.year.values if y < 2013],
            )
        ] = 1

        # and also Micro cars other than BEVs
        if "Micro" in self.array.coords["size"].values:
            self.array.loc[
                dict(
                    parameter="has_low_range",
                    powertrain=[
                        pt
                        for pt in [
                            "PHEV-e",
                            "PHEV-c-p",
                            "PHEV-c-d",
                            "FCEV",
                            "PHEV-p",
                            "PHEV-d",
                            "HEV-d",
                            "HEV-p",
                            "ICEV-p",
                            "ICEV-d",
                            "ICEV-g",
                        ]
                        if pt in self.array.coords["powertrain"].values
                    ],
                    size="Micro",
                )
            ] = 1

        # fill in NaNs due to `Micro` only existing for `BEV` powertrain
        if "Micro" in self.array.coords["size"].values:
            if "BEV" in self.array.powertrain:
                self.array.loc[dict(size="Micro")] = self.array.loc[
                    dict(size="Micro")
                ].fillna(1)
                self.array.loc[
                    dict(size="Micro", parameter="lifetime kilometers")
                ] = self.array.loc[
                    dict(
                        size="Micro", parameter="lifetime kilometers", powertrain="BEV"
                    )
                ]

    def set_battery_preferences(self, energy_storage):

        self.energy_storage = energy_storage or {
            "electric": {
                "type": "NMC-622",
            }
        }

        self["battery cell energy density"] = self[
            f"battery cell energy density, {self.energy_storage['electric']['type'].split('-')[0].strip()}"
        ]
        self["battery cell mass share"] = self[
            f"battery cell mass share, {self.energy_storage['electric']['type'].split('-')[0].strip()}"
        ]

    def adjust_cost(self) -> None:
        """
        This method adjusts costs of energy storage over time, to correct for the overly optimistic linear
        interpolation between years.

        """

        n_iterations = self.array.shape[-1]
        n_year = len(self.array.year.values)

        # If uncertainty is not considered, the cost factor equals 1.
        # Otherwise, a variability of +/-30% is added.

        if n_iterations == 1:
            cost_factor = 1
        else:
            if "reference" in self.array.value.values.tolist():
                cost_factor = np.ones((n_iterations, 1))
            else:
                cost_factor = np.random.triangular(0.7, 1, 1.3, (n_iterations, 1))

        # Correction of hydrogen tank cost, per kg
        # Correction of fuel cell stack cost, per kW
        if "FCEV" in self.array.powertrain:
            self.array.loc[
                dict(powertrain="FCEV", parameter="fuel tank cost per kg")
            ] = np.reshape(
                (1.078e58 * np.exp(-6.32e-2 * self.array.year.values) + 3.43e2)
                * cost_factor,
                (1, n_year, n_iterations),
            )

            self.array.loc[
                dict(powertrain="FCEV", parameter="fuel tank cost per kg")
            ] = np.reshape(
                (3.15e66 * np.exp(-7.35e-2 * self.array.year.values) + 2.39e1)
                * cost_factor,
                (1, n_year, n_iterations),
            )

        # Correction of energy battery system cost, per kWh
        list_batt = [
            i
            for i in ["BEV", "PHEV-e", "PHEV-c-p", "PHEV-c-d"]
            if i in self.array.powertrain
        ]
        if len(list_batt) > 0:
            self.array.loc[
                dict(powertrain=list_batt, parameter="energy battery cost per kWh")
            ] = np.reshape(
                (2.75e86 * np.exp(-9.61e-2 * self.array.year.values) + 5.059e1)
                * cost_factor,
                (1, 1, n_year, n_iterations),
            )

        # Correction of power battery system cost, per kW
        list_pwt = [
            i
            for i in [
                "ICEV-p",
                "ICEV-d",
                "ICEV-g",
                "PHEV-c-p",
                "PHEV-c-d",
                "FCEV",
                "HEV-p",
                "HEV-d",
            ]
            if i in self.array.powertrain
        ]

        if len(list_pwt) > 0:
            self.array.loc[
                dict(powertrain=list_pwt, parameter="power battery cost per kW")
            ] = np.reshape(
                (8.337e40 * np.exp(-4.49e-2 * self.array.year.values) + 11.17)
                * cost_factor,
                (1, 1, n_year, n_iterations),
            )

        # Correction of combustion powertrain cost for ICEV-g
        if "ICEV-g" in self.array.powertrain:
            self.array.loc[
                dict(powertrain="ICEV-g", parameter="combustion powertrain cost per kW")
            ] = np.clip(
                np.reshape(
                    (5.92e160 * np.exp(-0.1819 * self.array.year.values) + 26.76)
                    * cost_factor,
                    (1, n_year, n_iterations),
                ),
                None,
                100,
            )

    def adjust_fuel_mass(self) -> None:
        """
        This method adjusts the fuel mass over the years, to correct for the linear
        interpolation between years.

        """

        n_iterations = self.array.shape[-1]
        n_year = len(self.array.year.values)

        # Correction of hydrogen mass
        self.array.loc[dict(powertrain="FCEV", parameter="fuel mass")] = np.reshape(
            (1.078e58 * np.exp(-6.32e-2 * self.array.year.values) + 3.43e2),
            (1, 1, n_year, n_iterations),
        )

        # Correction of CNG mass

    def drop_hybrid(self) -> None:
        """
        This method drops the powertrains `PHEV-c-p`, `PHEV-c-d` and `PHEV-e` as they were only used to create the
        `PHEV` powertrain.
        :returns: Does not return anything. Modifies ``self.array`` in place.
        """
        self.array = self.array.sel(
            powertrain=[
                pt
                for pt in self.array.coords["powertrain"].values
                if pt not in ["PHEV-e", "PHEV-c-p", "PHEV-c-d"]
            ]
        )

    def set_electricity_consumption(self) -> None:
        """
        This method calculates the total electricity consumption for BEV and plugin-hybrid vehicles
        :returns: Does not return anything. Modifies ``self.array`` in place.
        """

        self["electricity consumption"] = (
            self["TtW energy"]
            / self["battery charge efficiency"]
            / 3600
            * (self["charger mass"] > 0)
        )

    def calculate_ttw_energy(self) -> None:
        """
        This method calculates the energy required to operate auxiliary services as well
        as to move the car. The sum is stored under the parameter label "TtW energy" in :attr:`self.array`.

        """

        self.energy = xr.DataArray(
            np.zeros(
                (
                    len(self.array.coords["size"]),
                    len(self.array.coords["powertrain"]),
                    4,
                    len(self.array.coords["year"]),
                    len(self.array.coords["value"]),
                    self.ecm.cycle.shape[0],
                )
            ).astype("float32"),
            coords=[
                self.array.coords["size"],
                self.array.coords["powertrain"],
                [
                    "auxiliary energy",
                    "motive energy",
                    "recuperated energy",
                    "negative motive energy",
                ],
                self.array.coords["year"],
                self.array.coords["value"],
                np.arange(self.ecm.cycle.shape[0]),
            ],
            dims=["size", "powertrain", "parameter", "year", "value", "second"],
        )

        self.energy.loc[dict(parameter="auxiliary energy")] = (
            self["auxiliary power demand"].values[..., None] / 1000
        )

        self.energy.loc[
            dict(
                parameter="auxiliary energy",
                powertrain=[
                    pt
                    for pt in [
                        "ICEV-d",
                        "ICEV-p",
                        "ICEV-g",
                        "PHEV-c-d",
                        "PHEV-c-p",
                        "HEV-p",
                        "HEV-d",
                    ]
                    if pt in self.array.coords["powertrain"].values
                ],
            )
        ] /= self.array.loc[
            dict(
                parameter="engine efficiency",
                powertrain=[
                    pt
                    for pt in [
                        "ICEV-d",
                        "ICEV-p",
                        "ICEV-g",
                        "PHEV-c-d",
                        "PHEV-c-p",
                        "HEV-p",
                        "HEV-d",
                    ]
                    if pt in self.array.coords["powertrain"].values
                ],
            )
        ].values[
            ..., None
        ]

        self.energy.loc[
            dict(
                parameter="auxiliary energy",
                powertrain=[
                    pt
                    for pt in ["FCEV"]
                    if pt in self.array.coords["powertrain"].values
                ],
            )
        ] /= self.array.sel(
            parameter="fuel cell system efficiency",
            powertrain=[
                pt for pt in ["FCEV"] if pt in self.array.coords["powertrain"].values
            ],
        ).values[
            ..., None
        ]

        motive_energy, recuperated_energy, distance = self.ecm.motive_energy_per_km(
            driving_mass=self["driving mass"],
            rr_coef=self["rolling resistance coefficient"],
            drag_coef=self["aerodynamic drag coefficient"],
            frontal_area=self["frontal area"],
            sizes=self.array.coords["size"].values,
            motor_power=self["electric power"],
        )

        self.energy.loc[dict(parameter="motive energy")] = np.clip(
            motive_energy / 1000, 0, None
        )

        self.energy.loc[dict(parameter="negative motive energy")] = (
            np.clip(motive_energy / 1000, None, 0) * -1
        )

        self.energy.loc[dict(parameter="motive energy")] /= self["TtW efficiency"]

        self.energy.loc[dict(parameter="motive energy")] = np.clip(
            self.energy.loc[dict(parameter="motive energy")],
            0,
            self["power"].values[..., None],
        )

        self.energy.loc[dict(parameter="recuperated energy")] = np.clip(
            recuperated_energy / 1000,
            self["power"].values[..., None] * -1,
            0,
        )

        self.energy.loc[dict(parameter="recuperated energy")] *= self[
            "recuperation efficiency"
        ].values[..., None]

        self.energy = self.energy.fillna(0)
        self.energy *= np.isfinite(self.energy)

        self["auxiliary energy"] = (
            self.energy.sel(parameter="auxiliary energy").sum(dim="second").T / distance
        ).T

        self["TtW energy"] = (
            self.energy.sel(
                parameter=["motive energy", "auxiliary energy", "recuperated energy"]
            )
            .sum(dim=["second", "parameter"])
            .T
            / distance
        ).T

        self["TtW energy, combustion mode"] = self["TtW energy"] * (
            self["combustion power share"] > 0
        )
        self["TtW energy, electric mode"] = self["TtW energy"] * (
            self["combustion power share"] == 0
        )

    def set_fuel_cell_parameters(self) -> None:
        """
        Specific setup for fuel cells, which are mild hybrids.
        Must be called after :meth:`.set_power_parameters`.
        """
        if "FCEV" in self.array.coords["powertrain"]:

            self.array.loc[
                dict(powertrain="FCEV", parameter="fuel cell system efficiency")
            ] = (
                self.array.loc[
                    dict(powertrain="FCEV", parameter="fuel cell stack efficiency")
                ]
                / self.array.loc[
                    dict(powertrain="FCEV", parameter="fuel cell own consumption")
                ]
            )
            self.array.loc[
                dict(powertrain="FCEV", parameter="fuel cell power share")
            ] = self.array.loc[
                dict(powertrain="FCEV", parameter="fuel cell power share")
            ].clip(
                min=0, max=1
            )
            self.array.loc[dict(powertrain="FCEV", parameter="fuel cell power")] = (
                self.array.loc[dict(powertrain="FCEV", parameter="power")]
                * self.array.loc[
                    dict(powertrain="FCEV", parameter="fuel cell power share")
                ]
                * self.array.loc[
                    dict(powertrain="FCEV", parameter="fuel cell own consumption")
                ]
            )
            # our basic fuel cell mass is based on a car fuel cell with 800 mW/cm2 and 0.51 kg/kW
            self.array.loc[
                dict(powertrain="FCEV", parameter="fuel cell stack mass")
            ] = (
                0.51
                * self.array.loc[dict(powertrain="FCEV", parameter="fuel cell power")]
                * (
                    800
                    / self.array.loc[
                        dict(
                            powertrain="FCEV", parameter="fuel cell power area density"
                        )
                    ]
                )
            )
            self.array.loc[
                dict(powertrain="FCEV", parameter="fuel cell ancillary BoP mass")
            ] = (
                self.array.loc[dict(powertrain="FCEV", parameter="fuel cell power")]
                * self.array.loc[
                    dict(
                        powertrain="FCEV",
                        parameter="fuel cell ancillary BoP mass per power",
                    )
                ]
            )
            self.array.loc[
                dict(powertrain="FCEV", parameter="fuel cell essential BoP mass")
            ] = (
                self.array.loc[dict(powertrain="FCEV", parameter="fuel cell power")]
                * self.array.loc[
                    dict(
                        powertrain="FCEV",
                        parameter="fuel cell essential BoP mass per power",
                    )
                ]
            )
            self.array.loc[
                dict(powertrain="FCEV", parameter="battery power")
            ] = self.array.loc[dict(powertrain="FCEV", parameter="fuel cell power")] * (
                1
                - self.array.loc[
                    dict(powertrain="FCEV", parameter="fuel cell power share")
                ]
            )
            self.array.loc[dict(powertrain="FCEV", parameter="battery cell mass")] = (
                self.array.loc[dict(powertrain="FCEV", parameter="battery power")]
                / self.array.loc[
                    dict(powertrain="FCEV", parameter="battery cell power density")
                ]
            )
            self.array.loc[
                dict(powertrain="FCEV", parameter="battery BoP mass")
            ] = self.array.loc[
                dict(powertrain="FCEV", parameter="battery cell mass")
            ] * (
                1
                - self.array.loc[
                    dict(powertrain="FCEV", parameter="battery cell mass share")
                ]
            )
            self.array.loc[
                dict(powertrain="FCEV", parameter="oxidation energy stored")
            ] = (
                self.array.loc[dict(powertrain="FCEV", parameter="fuel mass")]
                * 120
                / 3.6
            )  # kWh
            self.array.loc[dict(powertrain="FCEV", parameter="fuel tank mass")] = (
                self.array.loc[
                    dict(powertrain="FCEV", parameter="oxidation energy stored")
                ]
                * self.array.loc[
                    dict(powertrain="FCEV", parameter="H2 tank mass per energy")
                ]
            )

    def set_auxiliaries(self) -> None:
        """
        Calculates the power needed to operate the auxiliary services of the vehicle (heating, cooling).

        The demand for heat and cold are expressed as a fraction of the heating and cooling capacities

        .. note:

            Auxiliary power demand (W) = Base auxiliary power (W) +
            (Heating demand (dimensionless, between 0 and 1) * Heating power (W)) +
            (Cooling demand (dimensionless, between 0 and 1) * Cooling power (W))

        """
        self["auxiliary power demand"] = (
            self["auxilliary power base demand"]
            + self["heating thermal demand"] * self["heating energy consumption"]
            + self["cooling thermal demand"] * self["cooling energy consumption"]
        )

    def set_recuperation(self):
        self["recuperation efficiency"] = (
            self["drivetrain efficiency"] * self["battery charge efficiency"]
        )

    def set_battery_fuel_cell_replacements(self) -> None:
        """
        This methods calculates the fraction of the replacement battery needed to match the vehicle lifetime.

        .. note::
            if ``car lifetime`` = 200000 (km) and ``battery lifetime`` = 190000 (km) then ``replacement battery`` = 0.05

        .. note::
            It is debatable whether this is realistic or not. Car owners may not decide to invest in a new
            battery if the remaining lifetime of the car is only 10000 km. Also, a battery lifetime may be expressed
            in other terms, e.g., charging cycles.

        """
        # Here we assume that we can use fractions of a battery
        # (averaged across the fleet)
        self["battery lifetime replacements"] = finite(
            np.clip(
                (self["lifetime kilometers"] / self["battery lifetime kilometers"]) - 1,
                0,
                None,
            )
        )

        # The number of fuel cell replacements is based on the average distance driven
        # with a set of fuel cells given their lifetime expressed in hours of use.
        # The number is replacement is rounded *up* as we assume no allocation of burden
        # with a second life

        if "FCEV" in self.array.coords["powertrain"]:

            self.array.loc[
                dict(powertrain="FCEV", parameter="fuel cell lifetime replacements")
            ] = np.ceil(
                np.clip(
                    (
                        self.array.loc[
                            dict(powertrain="FCEV", parameter="lifetime kilometers")
                        ]
                        / (
                            self.ecm.cycle.sum(axis=0)
                            / self.ecm.cycle.shape[0]
                            * self.array.loc[
                                dict(
                                    powertrain="FCEV",
                                    parameter="fuel cell lifetime hours",
                                )
                            ].T
                        )
                    )
                    - 1,
                    0,
                    5,
                )
            )

    def set_car_masses(self) -> None:
        """
        Define ``curb mass``, ``driving mass``, and ``total cargo mass``.

            * `curb mass <https://en.wikipedia.org/wiki/Curb_weight>`__ is the mass of the vehicle and fuel, without people or cargo.
            * ``total cargo mass`` is the mass of the cargo and passengers.
            * ``driving mass`` is the ``curb mass`` plus ``total cargo mass``.

        .. note::
            driving mass = total cargo mass + driving mass

        """

        self["curb mass"] = self["glider base mass"] * (1 - self["lightweighting"])

        curb_mass_includes = [
            "fuel mass",
            "charger mass",
            "converter mass",
            "inverter mass",
            "power distribution unit mass",
            # Updates with set_components_mass
            "combustion engine mass",
            # Updates with set_components_mass
            "electric engine mass",
            # Updates with set_components_mass
            "powertrain mass",
            "fuel cell stack mass",
            "fuel cell ancillary BoP mass",
            "fuel cell essential BoP mass",
            "battery cell mass",
            "battery BoP mass",
            "fuel tank mass",
        ]
        self["curb mass"] += self[curb_mass_includes].sum(axis=2)

        self["total cargo mass"] = (
            self["average passengers"] * self["average passenger mass"]
            + self["cargo mass"]
        )
        self["driving mass"] = self["curb mass"] + self["total cargo mass"]

    def set_power_parameters(self) -> None:
        """Set electric and combustion motor powers based on input parameter ``power to mass ratio``."""
        # Convert from W/kg to kW
        self["power"] = self["power to mass ratio"] * self["curb mass"] / 1000
        self["combustion power share"] = self["combustion power share"].clip(
            min=0, max=1
        )
        self["combustion power"] = self["power"] * self["combustion power share"]
        self["electric power"] = self["power"] * (1 - self["combustion power share"])

    def set_component_masses(self) -> None:
        self["combustion engine mass"] = (
            self["combustion power"] * self["combustion mass per power"]
            + self["combustion fixed mass"]
        )
        self["electric engine mass"] = (
            self["electric power"] * self["electric mass per power"]
            + self["electric fixed mass"]
        ) * (self["electric power"] > 0)
        self["powertrain mass"] = (
            self["power"] * self["powertrain mass per power"]
            + self["powertrain fixed mass"]
        )

    def set_share_recuperated_energy(self) -> None:
        """Calculate the share of recuperated energy, over the total negative motive energy"""

        self["share recuperated energy"] = (
            self.energy.loc[dict(parameter="recuperated energy")].sum(dim="second") * -1
        ) / self.energy.loc[dict(parameter="negative motive energy")].sum(dim="second")

        if "PHEV-d" in self.array.powertrain:
            self.array.loc[
                dict(powertrain="PHEV-c-d", parameter="share recuperated energy")
            ] = self.array.loc[
                dict(powertrain="PHEV-e", parameter="share recuperated energy")
            ]

        if "PHEV-p" in self.array.powertrain:
            self.array.loc[
                dict(powertrain="PHEV-c-p", parameter="share recuperated energy")
            ] = self.array.loc[
                dict(powertrain="PHEV-e", parameter="share recuperated energy")
            ]

    def set_electric_utility_factor(self, uf: float = None) -> None:
        """Set the electric utility factor according to a sampled values in Germany (ICTT 2020)
        https://theicct.org/sites/default/files/publications/PHEV-white%20paper-sept2020-0.pdf

        Real-world range in simulation 20 km 30 km 40 km 50 km 60 km 70 km 80 km
        Observed UF for Germany (Sample-size weighted regression ± 2 standard errors)
        Observed UF private (in %) 30±2 41±2 50±3 58±3 65±3 71±3 75±3

        which correlated the share of km driven in electric-mode to the capacity of the battery
        (the range that can be driven in battery-depleting mode).

        The argument `uf` is used to override this relation, if needed.
        `uf` must be a ratio between 0 and .75, for each."""

        if "PHEV-e" in self.array.coords["powertrain"].values:
            if uf is None:
                self.array.loc[
                    dict(powertrain="PHEV-e", parameter="electric utility factor")
                ] = np.clip(
                    np.interp(
                        self.array.loc[dict(powertrain="PHEV-e", parameter="range")],
                        [0, 20, 30, 40, 50, 60, 70, 80],
                        [0, 0.3, 0.41, 0.5, 0.58, 0.65, 0.71, 0.75],
                    ),
                    0,
                    0.75,
                )
            else:
                self.array.loc[
                    dict(powertrain="PHEV-e", parameter="electric utility factor")
                ] = np.array(uf).reshape([1, -1, 1])

    def create_PHEV(self):
        """PHEV-p/d is the range-weighted average between PHEV-c-p/PHEV-c-d and PHEV-e."""

        if "PHEV-d" in self.array.coords["powertrain"].values:

            self.array.loc[:, "PHEV-d", :, :, :] = (
                self.array.loc[:, "PHEV-e", :, :, :]
                * self.array.loc[:, "PHEV-e", "electric utility factor", :, :]
            ) + (
                self.array.loc[:, "PHEV-c-d", :, :, :]
                * (1 - self.array.loc[:, "PHEV-e", "electric utility factor", :, :])
            )

            self.array.loc[
                dict(parameter="electric utility factor", powertrain=["PHEV-d"])
            ] = self.array.loc[
                dict(parameter="electric utility factor", powertrain=["PHEV-e"])
            ].values

            self.energy.loc[
                dict(
                    parameter=[
                        "motive energy",
                        "auxiliary energy",
                        "recuperated energy",
                    ],
                    powertrain=["PHEV-d"],
                )
            ] = (
                self.array.loc[
                    dict(parameter="electric utility factor", powertrain=["PHEV-e"])
                ]
                * self.energy.loc[
                    dict(
                        parameter=[
                            "motive energy",
                            "auxiliary energy",
                            "recuperated energy",
                        ],
                        powertrain=["PHEV-e"],
                    )
                ]
            ).values.transpose(
                0, 1, 4, 2, 3, 5
            ) + (
                (
                    1
                    - self.array.loc[
                        dict(parameter="electric utility factor", powertrain="PHEV-e")
                    ]
                )
                * self.energy.loc[
                    dict(
                        parameter=[
                            "motive energy",
                            "auxiliary energy",
                            "recuperated energy",
                        ],
                        powertrain=["PHEV-c-d"],
                    )
                ]
            ).values.transpose(
                0, 3, 4, 1, 2, 5
            )

            # We need to preserve the fuel mass and fuel tank mass
            self.array.loc[
                dict(
                    parameter=[
                        "fuel mass",
                        "fuel tank mass",
                        "oxidation energy stored",
                    ],
                    powertrain="PHEV-d",
                )
            ] = self.array.loc[
                dict(
                    parameter=[
                        "fuel mass",
                        "fuel tank mass",
                        "oxidation energy stored",
                    ],
                    powertrain="PHEV-c-d",
                )
            ]

            # we need to preserve the battery mass from PHEV-e as well
            self.array.loc[
                dict(
                    parameter=[
                        "energy battery mass",
                        "battery BoP mass",
                        "battery cell mass",
                        "battery DoD",
                        "battery cell energy density",
                    ],
                    powertrain="PHEV-d",
                )
            ] = self.array.loc[
                dict(
                    parameter=[
                        "energy battery mass",
                        "battery BoP mass",
                        "battery cell mass",
                        "battery DoD",
                        "battery cell energy density",
                    ],
                    powertrain="PHEV-e",
                )
            ]

            # We store the tank-to-wheel energy consumption
            # in combustion and electric mode separately
            self.array.loc[
                dict(parameter="TtW energy, combustion mode", powertrain="PHEV-d")
            ] = self.array.loc[dict(parameter="TtW energy", powertrain="PHEV-c-d")]

            self.array.loc[
                dict(parameter="TtW energy, electric mode", powertrain="PHEV-d")
            ] = self.array.loc[dict(parameter="TtW energy", powertrain="PHEV-e")]

            # We need to recalculate the range as well
            self.array.loc[dict(parameter="range", powertrain="PHEV-d")] = (
                self.array.loc[
                    dict(parameter="oxidation energy stored", powertrain="PHEV-d")
                ]
                * 3600
                / self.array.loc[
                    dict(parameter="TtW energy, combustion mode", powertrain="PHEV-d")
                ]
            )

            self.array.loc[dict(parameter="range", powertrain="PHEV-d")] += (
                self.array.loc[
                    dict(parameter="electric energy stored", powertrain="PHEV-d")
                ]
                * 3600
                / self.array.loc[
                    dict(parameter="TtW energy, electric mode", powertrain="PHEV-e")
                ]
            )

        if "PHEV-p" in self.array.coords["powertrain"].values:

            self.array.loc[:, "PHEV-p", :, :, :] = (
                self.array.loc[:, "PHEV-e", :, :, :]
                * self.array.loc[:, "PHEV-e", "electric utility factor", :, :]
            ) + (
                self.array.loc[:, "PHEV-c-p", :, :, :]
                * (1 - self.array.loc[:, "PHEV-e", "electric utility factor", :, :])
            )

            self.array.loc[
                dict(parameter="electric utility factor", powertrain=["PHEV-p"])
            ] = self.array.loc[
                dict(parameter="electric utility factor", powertrain=["PHEV-e"])
            ].values

            self.energy.loc[
                dict(
                    parameter=[
                        "motive energy",
                        "auxiliary energy",
                        "recuperated energy",
                    ],
                    powertrain=["PHEV-p"],
                )
            ] = (
                self.array.loc[
                    dict(parameter="electric utility factor", powertrain=["PHEV-e"])
                ]
                * self.energy.loc[
                    dict(
                        parameter=[
                            "motive energy",
                            "auxiliary energy",
                            "recuperated energy",
                        ],
                        powertrain=["PHEV-e"],
                    )
                ]
            ).values.transpose(
                0, 1, 4, 2, 3, 5
            ) + (
                (
                    1
                    - self.array.loc[
                        dict(parameter="electric utility factor", powertrain="PHEV-e")
                    ]
                )
                * self.energy.loc[
                    dict(
                        parameter=[
                            "motive energy",
                            "auxiliary energy",
                            "recuperated energy",
                        ],
                        powertrain=["PHEV-c-p"],
                    )
                ]
            ).values.transpose(
                0, 3, 4, 1, 2, 5
            )

            # We need to preserve the fuel mass and fuel tank mass
            self.array.loc[
                dict(
                    parameter=[
                        "fuel mass",
                        "fuel tank mass",
                        "oxidation energy stored",
                    ],
                    powertrain="PHEV-p",
                )
            ] = self.array.loc[
                dict(
                    parameter=[
                        "fuel mass",
                        "fuel tank mass",
                        "oxidation energy stored",
                    ],
                    powertrain="PHEV-c-p",
                )
            ]

            # we need to preserve the battery mass from PHEV-e as well
            self.array.loc[
                dict(
                    parameter=[
                        "energy battery mass",
                        "battery BoP mass",
                        "battery cell mass",
                        "battery DoD",
                        "battery cell energy density",
                    ],
                    powertrain="PHEV-p",
                )
            ] = self.array.loc[
                dict(
                    parameter=[
                        "energy battery mass",
                        "battery BoP mass",
                        "battery cell mass",
                        "battery DoD",
                        "battery cell energy density",
                    ],
                    powertrain="PHEV-e",
                )
            ]

            # We store the tank-to-wheel energy consumption
            # in combustion and electric mode separately
            self.array.loc[
                dict(parameter="TtW energy, combustion mode", powertrain="PHEV-p")
            ] = self.array.loc[dict(parameter="TtW energy", powertrain="PHEV-c-p")]

            self.array.loc[
                dict(parameter="TtW energy, electric mode", powertrain="PHEV-p")
            ] = self.array.loc[dict(parameter="TtW energy", powertrain="PHEV-e")]

            # We need to recalculate the range as well
            self.array.loc[dict(parameter="range", powertrain="PHEV-p")] = (
                self.array.loc[
                    dict(parameter="oxidation energy stored", powertrain="PHEV-p")
                ]
                * 3600
                / self.array.loc[
                    dict(parameter="TtW energy, combustion mode", powertrain="PHEV-p")
                ]
            )

            self.array.loc[dict(parameter="range", powertrain="PHEV-p")] += (
                self.array.loc[
                    dict(parameter="electric energy stored", powertrain="PHEV-p")
                ]
                * 3600
                / self.array.loc[
                    dict(parameter="TtW energy, electric mode", powertrain="PHEV-e")
                ]
            )

    def set_battery_properties(self) -> None:
        """
        Calculate mass and power of batteries.
        :return:
        """

        self.array.loc[dict(parameter="battery cell mass")] = (
            self.array.loc[dict(parameter="energy battery mass")]
            * self.array.loc[dict(parameter="battery cell mass share")]
        )

        self.array.loc[dict(parameter="battery BoP mass")] = self.array.loc[
            dict(parameter="energy battery mass")
        ] * (1 - self.array.loc[dict(parameter="battery cell mass share")])

    def set_range(self) -> None:
        """
        Calculate range autonomy of vehicles
        :return:
        """

        list_combustion = [
            i
            for i in [
                "ICEV-p",
                "HEV-p",
                "HEV-d",
                "PHEV-c-p",
                "PHEV-c-d",
                "ICEV-d",
                "ICEV-g",
                "FCEV",
            ]
            if i in self.array.powertrain
        ]

        list_electric = [i for i in ["BEV", "PHEV-e"] if i in self.array.powertrain]

        if len(list_combustion) > 0:

            fuel_mass = self.array.loc[
                dict(powertrain=list_combustion, parameter="fuel mass")
            ]  # in kg
            lhv = self.array.loc[
                dict(powertrain=list_combustion, parameter="LHV fuel MJ per kg")
            ]  # MJ/kg
            TtW = self.array.loc[
                dict(powertrain=list_combustion, parameter="TtW energy")
            ]  # kj/km
            self.array.loc[dict(powertrain=list_combustion, parameter="range")] = (
                fuel_mass * lhv * 1000
            ) / TtW  # -> km

        if len(list_electric) > 0:

            energy_stored = self.array.loc[
                dict(powertrain=list_electric, parameter="electric energy stored")
            ]  # in kWh
            battery_DoD = self.array.loc[
                dict(powertrain=list_electric, parameter="battery DoD")
            ]  # maximum depth of discharge allowed (80%)
            TtW = self.array.loc[
                dict(powertrain=list_electric, parameter="TtW energy")
            ]  # kj/km
            self.array.loc[dict(powertrain=list_electric, parameter="range")] = (
                energy_stored * battery_DoD * 3.6 * 1000 / TtW
            )  # -> km

    def set_energy_stored_properties(self) -> None:
        """
        Calculate size and capacity of onboard
        energy storage components.
        :return:
        """

        list_combustion = [
            i
            for i in [
                "ICEV-p",
                "HEV-p",
                "HEV-d",
                "PHEV-c-p",
                "PHEV-c-d",
                "ICEV-d",
            ]
            if i in self.array.powertrain
        ]

        if len(list_combustion) > 0:

            self.array.loc[
                dict(powertrain=list_combustion, parameter="oxidation energy stored")
            ] = (
                self.array.loc[dict(powertrain=list_combustion, parameter="fuel mass")]
                * self.array.loc[
                    dict(powertrain=list_combustion, parameter="LHV fuel MJ per kg")
                ]
                / 3.6
            )

            self.array.loc[
                dict(powertrain=list_combustion, parameter="fuel tank mass")
            ] = (
                self.array.loc[
                    dict(
                        powertrain=list_combustion, parameter="oxidation energy stored"
                    )
                ]
                * self.array.loc[
                    dict(
                        powertrain=list_combustion,
                        parameter="fuel tank mass per energy",
                    )
                ]
            )

        if "ICEV-g" in self.array.coords["powertrain"].values:

            self.array.loc[
                dict(powertrain="ICEV-g", parameter="oxidation energy stored")
            ] = (
                self.array.loc[dict(powertrain="ICEV-g", parameter="fuel mass")]
                * self.array.loc[
                    dict(powertrain="ICEV-g", parameter="LHV fuel MJ per kg")
                ]
                / 3.6
            )

            self.array.loc[dict(powertrain="ICEV-g", parameter="fuel tank mass")] = (
                self.array.loc[
                    dict(powertrain="ICEV-g", parameter="oxidation energy stored")
                ]
                * self.array.loc[
                    dict(powertrain="ICEV-g", parameter="CNG tank mass slope")
                ]
                + self.array.loc[
                    dict(powertrain="ICEV-g", parameter="CNG tank mass intercept")
                ]
            )

        if "BEV" in self.array.coords["powertrain"].values:

            self.array.loc[
                dict(powertrain=["BEV"], parameter="electric energy stored")
            ] = (
                self.array.loc[dict(powertrain=["BEV"], parameter="battery cell mass")]
                * self.array.loc[
                    dict(powertrain=["BEV"], parameter="battery cell energy density")
                ]
            )

        if "PHEV-e" in self.array.coords["powertrain"].values:

            self.array.loc[
                dict(powertrain="PHEV-e", parameter="electric energy stored")
            ] = (
                self.array.loc[dict(powertrain="PHEV-e", parameter="battery cell mass")]
                * self.array.loc[
                    dict(powertrain="PHEV-e", parameter="battery cell energy density")
                ]
            )

    def set_costs(self) -> None:
        """
        Calculate the different cost types.
        :return:
        """
        self["glider cost"] = (
            self["glider base mass"] * self["glider cost slope"]
            + self["glider cost intercept"]
        )
        self["lightweighting cost"] = (
            self["glider base mass"]
            * self["lightweighting"]
            * self["glider lightweighting cost per kg"]
        )
        self["electric powertrain cost"] = (
            self["electric powertrain cost per kW"] * self["electric power"]
        )
        self["combustion powertrain cost"] = (
            self["combustion power"] * self["combustion powertrain cost per kW"]
        )
        self["fuel cell cost"] = self["fuel cell power"] * self["fuel cell cost per kW"]
        self["power battery cost"] = (
            self["battery power"] * self["power battery cost per kW"]
        )
        self["energy battery cost"] = (
            self["energy battery cost per kWh"] * self["electric energy stored"]
        )
        self["fuel tank cost"] = self["fuel tank cost per kg"] * self["fuel mass"]
        # Per km
        self["energy cost"] = self["energy cost per kWh"] * self["TtW energy"] / 3600

        # For battery, need to divide cost of electricity
        # at battery by efficiency of charging
        # to get costs at the "wall socket".

        if "BEV" in self.array.coords["powertrain"].values:
            self.array.loc[
                dict(powertrain="BEV", parameter="energy cost")
            ] /= self.array.loc[
                dict(powertrain="BEV", parameter="battery charge efficiency")
            ]

        self["component replacement cost"] = (
            self["energy battery cost"] * self["battery lifetime replacements"]
            + self["fuel cell cost"] * self["fuel cell lifetime replacements"]
        )

        with open(DATA_DIR / "purchase_cost_params.yaml", "r") as stream:
            to_markup = yaml.safe_load(stream)["markup"]

        self[to_markup] *= self["markup factor"]

        # calculate costs per km:
        self["lifetime"] = self["lifetime kilometers"] / self["kilometers per year"]
        i = self["interest rate"]
        lifetime = self["lifetime"]
        amortisation_factor = ne.evaluate("i + (i / ((1 + i) ** lifetime - 1))")

        with open(DATA_DIR / "purchase_cost_params.yaml", "r") as stream:
            purchase_cost_params = yaml.safe_load(stream)["purchase"]

        self["purchase cost"] = self[purchase_cost_params].sum(axis=2)

        # per km
        self["amortised purchase cost"] = (
            self["purchase cost"] * amortisation_factor / self["kilometers per year"]
        )

        # per km
        self["maintenance cost"] = (
            self["maintenance cost per glider cost"]
            * self["glider cost"]
            / self["kilometers per year"]
        )

        # simple assumption that component replacement occurs at half of life.
        km_per_year = self["kilometers per year"]
        com_repl_cost = self["component replacement cost"]
        self["amortised component replacement cost"] = ne.evaluate(
            "(com_repl_cost * ((1 - i) ** lifetime / 2) * amortisation_factor / km_per_year)"
        )

        self["total cost per km"] = (
            self["energy cost"]
            + self["amortised purchase cost"]
            + self["maintenance cost"]
            + self["amortised component replacement cost"]
        )

    def set_ttw_efficiency(self) -> None:
        """
        Fill in the tank-to-wheel efficiency
        calculated by `calculate_ttw_efficiency`.
        :return:
        """
        _ = lambda array: np.where(array == 0, 1, array)

        self["TtW efficiency"] = (
            _(self["battery discharge efficiency"])
            * _(self["fuel cell system efficiency"])
            * self["drivetrain efficiency"]
            * self["engine efficiency"]
        )

    def set_hot_emissions(self) -> None:
        """
        Calculate hot pollutant emissions based on ``driving cycle``.
        The driving cycle is passed to the :class:`HotEmissionsModel` class
        and :meth:`get_emissions_per_powertrain`
        return emissions per substance per second of driving cycle.
        Those are summed up and divided by
        the distance driven, to obtain emissions, in kg per km.
        :return: Does not return anything. Modifies ``self.array`` in place.
        """
        hem = HotEmissionsModel(self.ecm.cycle, self.ecm.cycle_name)

        with open(DATA_DIR / "exhaust_flows.yaml", "r") as stream:
            list_direct_emissions = yaml.safe_load(stream)

        l_y = []
        for year in self.array.year.values:
            # European emission standards function of registration year
            # Here the registration year is synonymous of production year.
            if year < 1993:
                l_y.append(1)
            if 1993 <= year < 1997:
                l_y.append(1)
            if 1997 <= year < 2001:
                l_y.append(2)
            if 2001 <= year < 2006:
                l_y.append(3)
            if 2006 <= year < 2011:
                l_y.append(4)
            if 2011 <= year < 2015:
                l_y.append(5)
            if 2015 <= year < 2017:
                l_y.append(6.0)
            if 2017 <= year < 2019:
                l_y.append(6.1)
            if 2019 <= year < 2021:
                l_y.append(6.2)
            if year >= 2021:
                l_y.append(6.3)

        # to calculate emissions and degradation factors
        # we need the vehicle's lifetime, annual mileage
        # as well as its instant fuel consumption

        list_diesel = [
            i
            for i in ["ICEV-d", "PHEV-c-d", "HEV-d", "PHEV-d"]
            if i in self.array.powertrain
        ]

        if len(list_diesel) > 0:

            energy_consumption = self.energy.sel(
                powertrain=list_diesel,
                parameter=["motive energy", "auxiliary energy", "recuperated energy"],
            ).sum(dim="parameter")

            self.array.loc[
                dict(
                    powertrain=list_diesel,
                    parameter=list_direct_emissions,
                )
            ] = hem.get_hot_emissions(
                powertrain_type="ICEV-d",
                euro_class=l_y,
                lifetime_km=self.array.loc[
                    dict(powertrain=list_diesel, parameter="lifetime kilometers")
                ],
                energy_consumption=energy_consumption,
                yearly_km=self.array.loc[
                    dict(powertrain=list_diesel, parameter="kilometers per year")
                ],
            )

        list_petrol = [
            i
            for i in ["ICEV-p", "PHEV-c-p", "HEV-p", "PHEV-p"]
            if i in self.array.powertrain
        ]

        if len(list_petrol) > 0:

            energy_consumption = self.energy.sel(
                powertrain=list_petrol,
                parameter=["motive energy", "auxiliary energy", "recuperated energy"],
            ).sum(dim="parameter")

            self.array.loc[
                dict(
                    powertrain=list_petrol,
                    parameter=list_direct_emissions,
                )
            ] = hem.get_hot_emissions(
                powertrain_type="ICEV-p",
                euro_class=l_y,
                lifetime_km=self.array.loc[
                    dict(powertrain=list_petrol, parameter="lifetime kilometers")
                ],
                energy_consumption=energy_consumption,
                yearly_km=self.array.loc[
                    dict(powertrain=list_petrol, parameter="kilometers per year")
                ],
            )

        if "ICEV-g" in self.array.powertrain:

            energy_consumption = self.energy.sel(
                powertrain=["ICEV-g"],
                parameter=["motive energy", "auxiliary energy", "recuperated energy"],
            ).sum(dim="parameter")

            self.array.loc[
                dict(
                    powertrain=["ICEV-g"],
                    parameter=list_direct_emissions,
                )
            ] = hem.get_hot_emissions(
                powertrain_type=["ICEV-g"],
                euro_class=l_y,
                lifetime_km=self.array.loc[
                    dict(powertrain=["ICEV-g"], parameter="lifetime kilometers")
                ],
                energy_consumption=energy_consumption,
                yearly_km=self.array.loc[
                    dict(powertrain=["ICEV-g"], parameter="kilometers per year")
                ],
            )

    def set_particulates_emission(self) -> None:
        """
        Calculate the emission of particulates according to
        https://www.eea.europa.eu/ds_resolveuid/6USNA27I4D

        and further disaggregated in:
        https://doi.org/10.1016/j.atmosenv.2020.117886

        for:

        - brake wear
        - tire wear
        - road wear
        - re-suspended road dust

        by considering:

        - vehicle mass
        - driving situation (urban, rural, motorway)

        into the following fractions:

        - PM 2.5
        - PM 10

        Emissions are subdivided in compartments: urban, suburban and rural.

        """

        list_param = [
            "tire wear emissions",
            "brake wear emissions",
            "road wear emissions",
            "road dust emissions",
        ]

        pem = ParticulatesEmissionsModel(
            cycle_name=self.ecm.cycle_name,
            cycle=self.ecm.cycle,
            mass=self["driving mass"],
        )

        res = pem.get_abrasion_emissions()
        self[list_param] = res

        # brake emissions are discounted by the use of regenerative braking
        self["brake wear emissions"] *= 1 - self["share recuperated energy"]

    def set_noise_emissions(self) -> None:
        """
        Calculate noise emissions based on ``driving cycle``.
        The driving cycle is passed to the :class:`NoiseEmissionsModel` class and :meth:`get_sound_power_per_compartment`
        returns emissions per compartment type ("rural", "non-urban" and "urban") per second of driving cycle.

        Noise emissions are not differentiated by size classes at the moment, but only by powertrain "type"
        (e.g., combustion, hybrid and electric)

        :return: Does not return anything. Modifies ``self.array`` in place.
        """
        nem = NoiseEmissionsModel(self.ecm.cycle, self.ecm.cycle_name)

        with open(DATA_DIR / "noise_flows.yaml", "r") as stream:
            list_noise_emissions = yaml.safe_load(stream)

        list_pwt = [
            i
            for i in [
                "ICEV-p",
                "PHEV-c-p",
                "ICEV-g",
                "ICEV-d",
                "PHEV-c-d",
            ]
            if i in self.array.powertrain
        ]

        self.array.loc[
            dict(powertrain=list_pwt, parameter=list_noise_emissions)
        ] = nem.get_sound_power_per_compartment("combustion")

        list_pwt = [i for i in ["BEV", "FCEV", "PHEV-e"] if i in self.array.powertrain]

        self.array.loc[
            dict(powertrain=list_pwt, parameter=list_noise_emissions)
        ] = nem.get_sound_power_per_compartment("electric")

        list_pwt = [i for i in ["HEV-p", "HEV-d"] if i in self.array.powertrain]

        self.array.loc[
            dict(powertrain=list_pwt, parameter=list_noise_emissions)
        ] = nem.get_sound_power_per_compartment("hybrid")

    def calculate_cost_impacts(self, sensitivity=False, scope=None) -> xr.DataArray:
        """
        This method returns an array with cost values per vehicle-km,
        subdivided into the following groups:

            * Purchase
            * Maintenance
            * Component replacement
            * Energy
            * Total cost of ownership

        :return: A xarray array with cost information per vehicle-km
        """

        if scope is None:
            scope = {
                "size": self.array.coords["size"].values.tolist(),
                "powertrain": self.array.coords["powertrain"].values.tolist(),
                "year": self.array.coords["year"].values.tolist(),
            }

        list_cost_cat = [
            "purchase",
            "maintenance",
            "component replacement",
            "energy",
            "total",
        ]

        response = self.array.sel(
            powertrain=scope["powertrain"],
            size=scope["size"],
            year=scope["year"],
            parameter=[
                "amortised purchase cost",
                "maintenance cost",
                "amortised component replacement cost",
                "energy cost",
                "total cost per km",
            ],
        )

        response.coords["parameter"] = list_cost_cat

        if not sensitivity:
            return response

        return response / response.sel(value="reference")



"""
particulates_emissions.py contains ParticulatesEmissionsModel which calculated the amount of
abrasion particles emitted, given a driving cycle.
"""
import numpy as np
import xarray as xr


class ParticulatesEmissionsModel:
    """
    Calculate particulates emissions based on the method described in:
    https://www.eea.europa.eu/ds_resolveuid/6USNA27I4D

    and further disaggregated in:
    https://doi.org/10.1016/j.atmosenv.2020.117886

    Include emission from:

    - brake wear
    - tire wear
    - road wear
    - re-suspended road dust

    by considering:

    - vehicle mass
    - driving situation (urban, rural, motorway)

    into the following fractions:

    - PM 2.5
    - PM 10

    Emissions are subdivided in compartments: urban, suburban and rural.

    :param cycle: Driving cycle. Pandas Series of second-by-second speeds (km/h) or name (str)
        of cycle e.g., "Urban delivery", "Regional delivery", "Long haul".
    :param cycle_name: name of the driving cycle. Str.


    """

    def __init__(self, cycle: np.ndarray, cycle_name: str, mass: xr.DataArray) -> None:

        self.mass = mass.values / 1000  # in tons
        self.cycle = cycle
        self.cycle_name = cycle_name
        self.cycle_environment = {
            "WLTC": {
                "urban start": 0,
                "urban stop": 590,
                "suburban start": 591,
                "suburban stop": 1023,
                "rural start": 1024,
                "rural stop": 1801,
            },
            "WLTC 3.1": {"urban start": 0, "urban stop": 590},
            "WLTC 3.2": {"suburban start": 0, "suburban stop": 433},
            "WLTC 3.3": {"rural start": 0, "rural stop": 455},
            "WLTC 3.4": {"rural start": 0, "rural stop": 323},
            "CADC Urban": {"urban start": 0, "urban stop": 994},
            "CADC Road": {"suburban start": 0, "suburban stop": 1082},
            "CADC Motorway": {"rural start": 0, "rural stop": 1068},
            "CADC Motorway 130": {"rural start": 0, "rural stop": 1068},
            "CADC": {
                "urban start": 0,
                "urban stop": 994,
                "suburban start": 995,
                "suburban stop": 2077,
                "rural start": 2078,
                "rural stop": 3144,
            },
            "NEDC": {
                "urban start": 0,
                "urban stop": 780,
                "rural start": 781,
                "rural stop": 1180,
            },
            "custom": {"urban start": 0, "urban stop": cycle[0][-1]},
        }

    def get_abrasion_emissions(self) -> np.ndarray:

        (
            tire_pm10_urban,
            tire_pm10_rural,
            tire_pm10_motorway,
            tire_pm25_urban,
            tire_pm25_rural,
            tire_pm25_motorway,
        ) = self.get_tire_wear_emissions()

        (
            brake_pm10_urban,
            brake_pm10_rural,
            brake_pm10_motorway,
            brake_pm25_urban,
            brake_pm25_rural,
            brake_pm25_motorway,
        ) = self.get_brake_wear_emissions()

        road_pm10, road_pm25 = self.get_road_wear_emissions()
        dust_pm10, dust_pm25 = self.get_resuspended_road_dust()

        urban_start = self.cycle_environment[self.cycle_name].get("urban start", 0)
        urban_stop = self.cycle_environment[self.cycle_name].get("urban stop", 0)
        suburban_start = self.cycle_environment[self.cycle_name].get(
            "suburban start", 0
        )
        suburban_stop = self.cycle_environment[self.cycle_name].get("suburban stop", 0)
        rural_start = self.cycle_environment[self.cycle_name].get("rural start", 0)
        rural_stop = self.cycle_environment[self.cycle_name].get("rural stop", 0)

        cycle_length = self.cycle.shape[0]
        urban_share = (urban_stop - urban_start) / cycle_length
        suburban_share = (suburban_stop - suburban_start) / cycle_length
        rural_share = (rural_stop - rural_start) / cycle_length

        tire_wear = (tire_pm10_urban + tire_pm25_urban) * urban_share
        tire_wear += (tire_pm10_rural + tire_pm25_rural) * suburban_share
        tire_wear += (tire_pm10_motorway + tire_pm25_motorway) * rural_share

        brake_wear = (brake_pm10_urban + brake_pm25_urban) * urban_share
        brake_wear += (brake_pm10_rural + brake_pm25_rural) * suburban_share
        brake_wear += (brake_pm10_motorway + brake_pm25_motorway) * rural_share

        road_wear = road_pm10 + road_pm25
        road_dust = dust_pm10 + dust_pm25

        res = np.vstack(
            (
                tire_wear[None, ...],
                brake_wear[None, ...],
                road_wear[None, ...],
                road_dust[None, ...],
            )
        )

        return res.transpose(1, 2, 0, 3, 4)

    def get_tire_wear_emissions(self):
        """
        Returns tire wear emissions.

        :return:
        """

        # converted to kg per vkm
        pm10_urban = (5.8 * self.mass ** (1 / 2.3)) / 1e6
        pm25_urban = (8.2 * self.mass ** (1 / 2.3)) / 1e6

        pm10_rural = (4.5 * self.mass ** (1 / 2.3)) / 1e6
        pm25_rural = (6.4 * self.mass ** (1 / 2.3)) / 1e6

        pm10_motorway = (3.8 * self.mass ** (1 / 2.3)) / 1e6
        pm25_motorway = (5.5 * self.mass ** (1 / 2.3)) / 1e6

        return (
            pm10_urban,
            pm10_rural,
            pm10_motorway,
            pm25_urban,
            pm25_rural,
            pm25_motorway,
        )

    def get_brake_wear_emissions(self):
        """
        Returns brake wear emissions.

        :return:
        """
        # converted to kg per vkm
        pm10_urban = (4.2 * np.power(self.mass, 1 / 1.9)) / 1e6
        pm25_urban = (11 * np.power(self.mass, 1 / 1.9)) / 1e6

        pm10_rural = (1.8 * np.power(self.mass, 1 / 1.5)) / 1e6
        pm25_rural = (4.5 * np.power(self.mass, 1 / 1.5)) / 1e6

        pm10_motorway = (0.4 * np.power(self.mass, 1 / 1.3)) / 1e6
        pm25_motorway = (1.0 * np.power(self.mass, 1 / 1.3)) / 1e6

        return (
            pm10_urban,
            pm10_rural,
            pm10_motorway,
            pm25_urban,
            pm25_rural,
            pm25_motorway,
        )

    def get_road_wear_emissions(self):
        """
        Returns road wear emissions.

        :return:
        """
        # converted to kg per vkm
        pm10 = (2.8 * np.power(self.mass, 1 / 1.5)) / 1e6
        pm25 = (5.1 * np.power(self.mass, 1 / 1.5)) / 1e6

        return (pm10, pm25)

    def get_resuspended_road_dust(self):
        """
        Returns re-suspended road dust emissions.

        :return:
        """
        # converted to kg per vkm
        pm10 = (2 * np.power(self.mass, 1 / 1.1)) / 1e6
        pm25 = (8.2 * np.power(self.mass, 1 / 1.1)) / 1e6

        return (pm10, pm25)
